/*
 * AirCushionGeometry
 * Contains geometry of the air cushion
 * 
 * This federate uses its own axis system, with origin at the longitudinal center of the cushion, CL and BL
 * All distances and coordinates (except the global data) are given in this system
 * 
 * Made by Trygve Espeland Halvorsen, Spring 2008
 */

package no.marintek.simulator.ses;

import no.marintek.tools.dataobjects.Motion;
import no.marintek.simulator.vessel.ShipMotionConversion;
import no.marintek.simulator.environment.WavesModel;
import java.io.*;
import java.util.*;

public class AirCushionGeometry {
    int minSections = 40;       // Minimum number of nodes in sections
    
    // Common data
    public double L;            // cushion length
    public int sectionNumber;   // numbers of sections defining the air cushion
    public int[] coordNumber;   // Numbering following the 0, 1, 2, .. format. Numbered from stern seal and forward
    public double centerCushionx = 0;
    
    // Local data
    public double[] localxCoord;
    public double[][] localyCoord;
    public double[][] localzCoord;
    
    // Global data
    public double[][] globalStarBxCoord;
    public double[][] globalStarByCoord;
    public double[][] globalStarBzCoord;  
    public double[][] globalAPortxCoord;
    public double[][] globalAPortyCoord;
    public double[][] globalAPortzCoord; 
    public double[] globaldL;
    
    // Information about where the hull crosses the water surface (inside the air cushion). Global coordinates
    public double[] crossxCoordGlobalStarB;
    public double[] crossyCoordGlobalStarB;
    public double[] crosszCoordGlobalStarB;
    public double[] crossyCoordLocalStarB;
    public double[] crosszCoordLocalStarB;
    public int[] firstCoordInAirStarb;
    public double[] firstWetLengthStarb;
    
    // Information about where the hull crosses the water surface (inside the air cushion). Aport
    public double[] crossxCoordGlobalAPort;
    public double[] crossyCoordGlobalAPort;
    public double[] crosszCoordGlobalAPort;
    public double[] crossyCoordLocalAPort;
    public double[] crosszCoordLocalAPort;
    public int[] firstCoordInAirAPort;
    public double[] firstWetLengthAPort;
    
    public double[] globalShipDirectionVector = new double[3];
    
    // Element data
    public double[] dx;
    public double[][] elementLength;
    public double[][] yNormalVector;
    public double[][] zNormalVector;
    public double[] MomentArmx;
    public double[][] MomentArmy;
    public double[][] MomentArmz;
    
    
    public AirCushionGeometry(String filePath, double cushionxCoordStern, double cushionxCoordBow) {
        this.readmgfFile(filePath, cushionxCoordStern, cushionxCoordBow);
        this.defineElements();
    }
 
    
    public void updateStatic(Constants constant, double p_c, double draught) {
        this.findSubmergenceStatic(p_c, constant, draught);
    }
    
    
    public void updateGlobal(Motion cgMotion, ShipMotionConversion convertShipMotion, Constants constant, double p_c, double time) {
        // Update glob coordinates also where on the cushion the water crosses.
        // Ship direction
        // This is needed for leakage and integrating dry area
        this.makeGlobalCoord(cgMotion, convertShipMotion);
        this.makeShipDirectionVector(cgMotion, convertShipMotion);
        this.findSubmergence(p_c, constant, time);
        
    }
    
    private void initializeArrays() {
        coordNumber = new int[sectionNumber];
        
        localxCoord = new double[sectionNumber];
        localyCoord = new double[sectionNumber][minSections];
        localzCoord = new double[sectionNumber][minSections];
        
        globalStarBxCoord = new double[sectionNumber][minSections];
        globalStarByCoord = new double[sectionNumber][minSections];
        globalStarBzCoord = new double[sectionNumber][minSections];  
        globalAPortxCoord = new double[sectionNumber][minSections];
        globalAPortyCoord = new double[sectionNumber][minSections];
        globalAPortzCoord = new double[sectionNumber][minSections];
        globaldL  = new double[sectionNumber];
        
        crossxCoordGlobalStarB = new double[sectionNumber];
        crossyCoordGlobalStarB = new double[sectionNumber];
        crosszCoordGlobalStarB = new double[sectionNumber];
        crossyCoordLocalStarB = new double[sectionNumber];
        crosszCoordLocalStarB = new double[sectionNumber];
        firstCoordInAirStarb = new int[sectionNumber];     
        firstWetLengthStarb = new double[sectionNumber];
        
        crossxCoordGlobalAPort = new double[sectionNumber];
        crossyCoordGlobalAPort = new double[sectionNumber];
        crosszCoordGlobalAPort = new double[sectionNumber];
        crossyCoordLocalAPort = new double[sectionNumber];
        crosszCoordLocalAPort = new double[sectionNumber];
        firstCoordInAirAPort = new int[sectionNumber]; 
        firstWetLengthAPort = new double[sectionNumber];
        
        dx = new double[sectionNumber];
        elementLength = new double[sectionNumber][minSections-1];
        yNormalVector = new double[sectionNumber][minSections-1];
        zNormalVector = new double[sectionNumber][minSections-1];
        MomentArmx= new double[sectionNumber];
        MomentArmy = new double[sectionNumber][minSections-1];
        MomentArmz = new double[sectionNumber][minSections-1];
    } 

    private void readmgfFile(String filePath, double cushionxCoordStern, double cushionxCoordBow) {
        System.out.println("SESFederate: Reading hull geometry file: " + filePath);

        // Local variables
        double Lpp = 0;
        double mgfxCoordStern = 0;  // Coordinate of aft end of air cushion, mgf coordinate system
        double mgfxCoordBow = 0;    // Coordinate of fore end of air cushion, mgf coordinate system
        int mgfsectionNumber = 0;
        int[] mgfcoordNumber = new int[1];
        double[] mgflocalxCoord = new double[1];
        double[][] mgflocalyCoord  = new double[1][1];
        double[][] mgflocalzCoord  = new double[1][1];
        int[] mgflowestCoordNr = new int[1];

        this.L = cushionxCoordBow - cushionxCoordStern; //=30
        this.centerCushionx = 0.5*(cushionxCoordBow + cushionxCoordStern); //=15

        try {
            BufferedReader input = new BufferedReader(new FileReader(filePath));
            
            // Read toptext. This text is not of interest
            try{
                for (int i=0; i<4; i++){
                    String dummy = input.readLine();
                }
            } catch (EOFException eof){
            }            
            
            // Read lpp
            String dummy = input.readLine();
            dummy = dummy.trim();
            Lpp = Double.valueOf(dummy).doubleValue();
            
            // Find cushion coordinates in mgf axis system
            mgfxCoordStern = Lpp/2-cushionxCoordStern;
            mgfxCoordBow = Lpp/2-cushionxCoordBow;
            
            // Read sectionNr
            String sectionDummy = "0";
            try{
                sectionDummy = input.readLine();
            } catch (EOFException eof){
                input.close();
            }

            mgfsectionNumber = Integer.parseInt(sectionDummy);
            // Initialize local variables
            mgfcoordNumber = new int[mgfsectionNumber];
            mgflocalxCoord = new double[mgfsectionNumber];
            mgflocalyCoord = new double[mgfsectionNumber][minSections];
            mgflocalzCoord = new double[mgfsectionNumber][minSections];
            mgflowestCoordNr = new int[mgfsectionNumber];
            
            // Read remaining data
            try{
                int i = 0;
                boolean running = true;
                do{
                    if (i!=0)
                        sectionDummy = input.readLine();
                    if (sectionDummy == null)
                        running = false;
                    String xCoordDummy = input.readLine();                   
                    String coordDummy = input.readLine(); 
                    
                    if (coordDummy != null) {
                        int sectionNr = Integer.parseInt(sectionDummy);
                        int coordNr =  Integer.parseInt(coordDummy);
 
                        // Store the x-coordinate of the section
                        mgflocalxCoord[sectionNr-1] = Double.valueOf(xCoordDummy).doubleValue();
                        
                        // Store number of coordinates in the section 
                        mgfcoordNumber[sectionNr-1] = coordNr;
                        
                        // Read and store coordinates
                        for (int j=0; j<coordNr; j++){
                            String yzDummy = input.readLine(); yzDummy = yzDummy.trim();
                            {
                                StringTokenizer dummyToken = new StringTokenizer(yzDummy);
                                mgflocalyCoord[sectionNr-1][j] = Double.valueOf(dummyToken.nextToken()).doubleValue();
                                mgflocalzCoord[sectionNr-1][j] = Double.valueOf(dummyToken.nextToken()).doubleValue();
                            }
                        }
                        // Find the lowest coordinate on the section
                        // If several has same z-coordinate - fint the one closest to CL
                        int candidate = 0;
                        for (int j=1; j<coordNr; j++) {
                            if (mgflocalzCoord[sectionNr-1][candidate] > mgflocalzCoord[sectionNr-1][j])
                                candidate = j;
                            else if (mgflocalzCoord[sectionNr-1][candidate] == mgflocalzCoord[sectionNr-1][j]
                                    && mgflocalyCoord[sectionNr-1][candidate] > mgflocalyCoord[sectionNr-1][j])
                                candidate = j;
                        }
                        mgflowestCoordNr[sectionNr-1] = candidate;
                        i++;
                    }
                } while (running == true);
            } catch (EOFException eof){
                input.close();
            }                                
        } catch (IOException e) {
            System.err.println("Error: " + e.toString());
        }    

        // Find the sections closest to cushion seals, and start- and end-sections
        int sealSternSection = 0;
        int sealBowSection = 0;
        int sternSection = mgfsectionNumber;
        int bowSection = 0;
        for (int sectionNr=1; sectionNr<mgfsectionNumber; sectionNr++) {
            if (Math.abs(mgfxCoordStern-mgflocalxCoord[sectionNr]) < Math.abs(mgfxCoordStern-mgflocalxCoord[sealSternSection]))
                sealSternSection = sectionNr;
            if (Math.abs(mgfxCoordBow-mgflocalxCoord[sectionNr]) < Math.abs(mgfxCoordBow-mgflocalxCoord[sealBowSection]))
                sealBowSection = sectionNr;
            if (mgflocalxCoord[sectionNr]<mgfxCoordStern && sectionNr<sternSection)
                sternSection = sectionNr;
            if (mgflocalxCoord[sectionNr]>mgfxCoordBow && sectionNr>bowSection)
                bowSection = sectionNr;
        }
        
        // Number of sections equals sections between seals + one for each seal
        this.sectionNumber = (bowSection-sternSection) + 3;
        
        this.initializeArrays();
        
        // Store cushion geometry
        // Geometry at stern seal
        this.coordNumber[0] = mgfcoordNumber[sealSternSection]-mgflowestCoordNr[sealSternSection];
        this.localxCoord[0] = -this.L/2;
        for (int coordNr=0; coordNr<this.coordNumber[0]; coordNr++) {
            int mgfCoordNr = mgflowestCoordNr[sealSternSection]+coordNr;
            this.localyCoord[0][coordNr] = mgflocalyCoord[sealSternSection][mgfCoordNr];
            this.localzCoord[0][coordNr] = mgflocalzCoord[sealSternSection][mgfCoordNr];            
        }
        // Geometry at bow seal
        this.coordNumber[this.sectionNumber-1] = mgfcoordNumber[sealBowSection] - mgflowestCoordNr[sealBowSection];
        this.localxCoord[this.sectionNumber-1] = this.L/2;
        for (int coordNr=0; coordNr<this.coordNumber[0]; coordNr++) {
            int mgfCoordNr = mgflowestCoordNr[sealBowSection]+coordNr;
            this.localyCoord[this.sectionNumber-1][coordNr] = mgflocalyCoord[sealBowSection][mgfCoordNr];
            this.localzCoord[this.sectionNumber-1][coordNr] = mgflocalzCoord[sealBowSection][mgfCoordNr];            
        }
        // Section between seals
        for (int sectionNr=1; sectionNr<this.sectionNumber-1; sectionNr++) {
            int mgfSection = sternSection + sectionNr - 1;
            this.coordNumber[sectionNr] = mgfcoordNumber[mgfSection] - mgflowestCoordNr[mgfSection];
            this.localxCoord[sectionNr] = -this.L/2 + mgfxCoordStern - mgflocalxCoord[mgfSection];
            for (int coordNr=0; coordNr<this.coordNumber[sectionNr]; coordNr++) {
                int mgfCoordNr = mgflowestCoordNr[mgfSection]+coordNr;
                this.localyCoord[sectionNr][coordNr] = mgflocalyCoord[mgfSection][mgfCoordNr];
                this.localzCoord[sectionNr][coordNr] = mgflocalzCoord[mgfSection][mgfCoordNr];
            }
        }
    }
    
    
    private void defineElements() {
        double y1;
        double z1;
        double y2;
        double z2;
        double lengthSecond;
        double length;
        for (int secNr=0; secNr<this.sectionNumber; secNr++) {
            for (int coordNr=1; coordNr<this.coordNumber[secNr]; coordNr++) {
                y1 = this.localyCoord[secNr][coordNr-1];
                z1 = this.localzCoord[secNr][coordNr-1];
                y2 = this.localyCoord[secNr][coordNr];
                z2 = this.localzCoord[secNr][coordNr];

                lengthSecond = (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1);
                length = Math.sqrt(lengthSecond);
                this.elementLength[secNr][coordNr-1] = length;
                
                // Positive direction of normal vectores are away from the hull
                this.yNormalVector[secNr][coordNr-1] = -(z2-z1)/length;
                this.zNormalVector[secNr][coordNr-1] = (y2-y1)/length;
                this.MomentArmy[secNr][coordNr-1] = 0.5*(y2+y1);
                this.MomentArmz[secNr][coordNr-1] = 0.5*(z2+z1);
            }   

            if (secNr == 0) {
                // Stern
                this.dx[secNr] = this.L/2 - 0.5*Math.abs(this.localxCoord[0]+this.localxCoord[1]);
                this.MomentArmx[secNr] = -this.L/2 + this.dx[secNr]/2;
            } else if (secNr == this.sectionNumber-1) {
                // Bow
                this.dx[secNr] = this.L/2 - 0.5*Math.abs(this.localxCoord[secNr]+this.localxCoord[secNr-1]);
                this.MomentArmx[secNr] = this.L/2 - this.dx[secNr]/2;
            } else {
                // Between stern and bow
                this.dx[secNr] = 0.5*Math.abs(this.localxCoord[secNr+1]-this.localxCoord[secNr-1]);
                this.MomentArmx[secNr] = 0.5*(this.localxCoord[secNr+1]+this.localxCoord[secNr-1]);
            }
        } 
    }
    
    
    private void makeGlobalCoord(Motion cgMotion, ShipMotionConversion convertShipMotion) {
   
        Motion pointMotion = new Motion();
        for (int secNr=0; secNr<this.sectionNumber; secNr++) {

            // ----- Starboard side -----
            for (int coordNr=0; coordNr<this.coordNumber[secNr]; coordNr++) {                     
                pointMotion = convertShipMotion.getMotionAtPoint(
                        this.localxCoord[secNr]+this.centerCushionx,
                        this.localyCoord[secNr][coordNr],
                        this.localzCoord[secNr][coordNr], cgMotion);
                this.globalStarBxCoord[secNr][coordNr] = pointMotion.ned.north; // Setting global x-coordinate of section
                this.globalStarByCoord[secNr][coordNr] = pointMotion.ned.east;  // Setting global y-coordinate of section
                this.globalStarBzCoord[secNr][coordNr] = pointMotion.ned.down;  // Setting global z-coordinate of section
            }    
            
            // ----- Aport side -----
            for (int coordNr=0; coordNr<this.coordNumber[secNr]; coordNr++) {  
                pointMotion = convertShipMotion.getMotionAtPoint( 
                        this.localxCoord[secNr]+this.centerCushionx,
                        -this.localyCoord[secNr][coordNr],
                        this.localzCoord[secNr][coordNr], cgMotion);
                this.globalAPortxCoord[secNr][coordNr] = pointMotion.ned.north; // Setting global x-coordinate of section
                this.globalAPortyCoord[secNr][coordNr] = pointMotion.ned.east;  // Setting global y-coordinate of section
                this.globalAPortzCoord[secNr][coordNr] = pointMotion.ned.down;  // Setting global z-coordinate of section
            }
        }  
        
        // ----- global dL -----
        // the global DL corresponds to local dx. When having pitch, 
        // the width of sections used to calculate air cushion volume is too large. 
        // Therefore the global version is used instead.
        Double DoubleNumber = Math.floor(this.sectionNumber*0.5);
        int midSectionNumber = DoubleNumber.intValue();
        double deltaX = 0.5*Math.abs(this.globalAPortxCoord[midSectionNumber+1][0]-this.globalAPortxCoord[midSectionNumber-1][0]);
        double deltaY = 0.5*Math.abs(this.globalAPortyCoord[midSectionNumber+1][0]-this.globalAPortyCoord[midSectionNumber-1][0]);
        double dxScale = Math.sqrt(deltaX*deltaX + deltaY*deltaY)
                /this.dx[midSectionNumber];
        for (int secNr=0; secNr<this.sectionNumber; secNr++) {
            this.globaldL[secNr] = this.dx[secNr]*dxScale;
        }
    }
    
    private void makeShipDirectionVector(Motion cgMotion, ShipMotionConversion convertShipMotion) {
        
        // ----- the ship's direction vector, normalized -----
        // This vector follows the global NED system
        // It is needed to obtain correct volume and leakage calculations
        double[] y = new double[3];
        double[] z = new double[3];
        double[] n = new double[3];
        Motion centerMotion = new Motion();
        Motion yMotion = new Motion();
        Motion zMotion = new Motion();

        centerMotion = convertShipMotion.getMotionAtPoint(0, 0, 0, cgMotion);
        yMotion = convertShipMotion.getMotionAtPoint(0, -1, 0, cgMotion);
        zMotion = convertShipMotion.getMotionAtPoint(0, 0, 1, cgMotion);

        y[0] = yMotion.ned.north-centerMotion.ned.north; 
        y[1] = yMotion.ned.east-centerMotion.ned.east; 
        y[2] = yMotion.ned.down-centerMotion.ned.down; 
        z[0] = zMotion.ned.north-centerMotion.ned.north; 
        z[1] = zMotion.ned.east-centerMotion.ned.east; 
        z[2] = zMotion.ned.down-centerMotion.ned.down;
        
        n[0] = y[1]*z[2] - y[2]*z[1];
        n[1] = y[2]*z[0] - y[0]*z[2];
        n[2] = y[0]*z[1] - y[1]*z[0];
        
        double length = Math.sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
        this.globalShipDirectionVector[0] = n[0]/length;
        this.globalShipDirectionVector[1] = n[1]/length;
        this.globalShipDirectionVector[2] = n[2]/length;
    }
    
    
    private void findSubmergenceStatic(double p_c, Constants constant, double draught) {
        
        double pressureWC = p_c/(constant.rho*constant.g);
        
        for (int secNr=0; secNr<this.sectionNumber; secNr++){

            // ---------------------------------------------
            // Find the two coordinates on section where hull crosses water surface
            // ---------------------------------------------          
            double y_cross1;
            double z_cross1;
            double y_cross2;
            double z_cross2;
            int areaNr1;
            int areaNr2;
            
            boolean goon = true;
            double submergence = 0; // Positive means above water
            double previousSubmergence = 0;
            int candidate = 0;

            // positive z is up
            while (goon) {                
                previousSubmergence = submergence;
                submergence = this.localzCoord[secNr][candidate] - draught + pressureWC;

                if (submergence > 0) { // If the point is above water - stop loop                   
                    goon = false; }
                else if (candidate == (this.coordNumber[secNr]-1)) {    // If the point is the last point (at the center line) - stop loop
                    goon = false; }            
                candidate++;
            } 
            candidate--;
            areaNr1 = candidate;
            
            // If entire section submerged or above water
            if (candidate == 0
                    || ((candidate==(this.coordNumber[secNr]-1)) && (submergence < 0))){
                //System.out.println("Case 1, leakage or totally submerged");
                y_cross1 = this.localyCoord[secNr][candidate];
                z_cross1 = this.localzCoord[secNr][candidate];

            } else {
                // Interpolate to find point where water surface crosses hull
                z_cross1 = this.localzCoord[secNr][candidate] - submergence;
                y_cross1 = this.localyCoord[secNr][candidate-1] +
                        (z_cross1 - this.localzCoord[secNr][candidate-1])
                        * (this.localyCoord[secNr][candidate] - this.localyCoord[secNr][candidate-1])
                        / (this.localzCoord[secNr][candidate] - this.localzCoord[secNr][candidate-1]);
            }

            // Find crossing point on opposite side of CL
            y_cross2 = -y_cross1;
            z_cross2 = z_cross1;
            areaNr2 = areaNr1;               
            
            // Store crossing data
            this.crossyCoordLocalStarB[secNr] = y_cross1;
            this.crosszCoordLocalStarB[secNr] = z_cross1;
            this.firstCoordInAirStarb[secNr] = areaNr1;
            this.crossyCoordLocalAPort[secNr] = y_cross2;
            this.crosszCoordLocalAPort[secNr] = z_cross2;
            this.firstCoordInAirAPort[secNr] = areaNr2;
            
            // Find wetted length of first element that touches water

            // ----- Starboard side ----- 
            if (areaNr1 == 0) {
                // If entire section below water
                this.firstWetLengthStarb[secNr] = 0; // Entire element length is wet - is included in the general wet length calculation
            } else {
                double y = y_cross1-this.localyCoord[secNr][areaNr1-1];
                double z = z_cross1-this.localzCoord[secNr][areaNr1-1];
                this.firstWetLengthStarb[secNr] = Math.sqrt(y*y + z*z);
            }
            
            // ----- Aport side -----  
            if (areaNr2 == 0) {
                // If entire section below water
                this.firstWetLengthAPort[secNr] = 0; // Entire element length is wet - is included in the general wet length calculation
            } else {
                double y = y_cross2+this.localyCoord[secNr][areaNr2-1];
                double z = z_cross2-this.localzCoord[secNr][areaNr2-1];
                this.firstWetLengthAPort[secNr] = Math.sqrt(y*y + z*z);
            }
            
        }
    }
            
    
    private void findSubmergence(double p_c, Constants constant, double time) {
        
        double pressureWC = p_c/(constant.rho*constant.g);
        WavesModel waves = new WavesModel();
        
        for (int secNr=0; secNr<this.sectionNumber; secNr++){

            // ---------------------------------------------
            // Find the two coordinates on hull where hull crosses water surface
            // ---------------------------------------------         
            
            // ----- Starboard side ----- 
            double x_cross1;
            double y_cross1;
            double z_cross1;
            double x_cross2;
            double y_cross2;
            double z_cross2;
            int areaNr1;
            int areaNr2;
            
            boolean goon = true;
            double submergence = 0; // Positive means above water
            double previousSubmergence = 0;
            double wave_elevation_xy;
            int candidate = -1;
            
            while (goon) {  
                candidate++;
                previousSubmergence = submergence;
                wave_elevation_xy = waves.getInterpolatedWaveElevation(time, this.globalStarBxCoord[secNr][candidate], this.globalStarByCoord[secNr][candidate])
                        + pressureWC;   // pos. down
                submergence = -this.globalStarBzCoord[secNr][candidate]+wave_elevation_xy;  // Positive means point above water
                if (submergence > 0)    // If the point is above water - stop loop                   
                    goon = false;
                else if (candidate == this.coordNumber[secNr]-1)
                    goon = false;            
            } 
            areaNr1 = candidate;
            
            // If entire section submerged or above water
            if (candidate == 0
                    || ((candidate==(this.coordNumber[secNr]-1)) && (submergence < 0))){
                x_cross1 = this.globalStarBxCoord[secNr][candidate];
                y_cross1 = this.globalStarByCoord[secNr][candidate];
                z_cross1 = this.globalStarBzCoord[secNr][candidate];
            } else {
                // Interpolate to find point where water surface crosses hull
                x_cross1 = this.globalStarBxCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalStarBxCoord[secNr][candidate]-this.globalStarBxCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
                y_cross1 = this.globalStarByCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalStarByCoord[secNr][candidate]-this.globalStarByCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
                z_cross1 = this.globalStarBzCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalStarBzCoord[secNr][candidate]-this.globalStarBzCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
            }
            
            // ----- Aport side ----- 
            goon = true;
            submergence = 0;    // Positive means above water
            previousSubmergence = 0;
            candidate = -1;
            while (goon) {  
                candidate++;
                previousSubmergence = submergence;
                wave_elevation_xy = waves.getInterpolatedWaveElevation(time, this.globalAPortxCoord[secNr][candidate], this.globalAPortyCoord[secNr][candidate])
                        + pressureWC;   // pos. down
                submergence = -this.globalAPortzCoord[secNr][candidate]+wave_elevation_xy;  // Positive means point above water
                if (submergence > 0)    // If the point is above water - stop loop                   
                    goon = false;
                else if (candidate == this.coordNumber[secNr]-1)
                    goon = false;            
            } 
            areaNr2 = candidate;

            // If entire section submerged or above water
            if (candidate == 0
                    || ((candidate==(this.coordNumber[secNr]-1)) && (submergence < 0))){
                x_cross2 = this.globalAPortxCoord[secNr][candidate];
                y_cross2 = this.globalAPortyCoord[secNr][candidate];
                z_cross2 = this.globalAPortzCoord[secNr][candidate];
            } else {              
                // Interpolate to find point where water surface crosses hull
                x_cross2 = this.globalAPortxCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalAPortxCoord[secNr][candidate]-this.globalAPortxCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
                y_cross2 = this.globalAPortyCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalAPortyCoord[secNr][candidate]-this.globalAPortyCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
                z_cross2 = this.globalAPortzCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalAPortzCoord[secNr][candidate]-this.globalAPortzCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
            }                                 

            // Store crossing data
            this.crossxCoordGlobalStarB[secNr] = x_cross1;
            this.crossyCoordGlobalStarB[secNr] = y_cross1;
            this.crosszCoordGlobalStarB[secNr] = z_cross1;
            this.firstCoordInAirStarb[secNr] = areaNr1;
            
            this.crossxCoordGlobalAPort[secNr] = x_cross2;
            this.crossyCoordGlobalAPort[secNr] = y_cross2;
            this.crosszCoordGlobalAPort[secNr] = z_cross2;
            this.firstCoordInAirAPort[secNr] = areaNr2;
    
            
            // Also need crossing coordinates in local coordinates
            // These are needed as forces are calculated using the local coordinates
            
            double crossy1 = 0;
            double crossy2 = 0;
            double crossz1 = 0;
            double crossz2 = 0;

            // ----- Starboard side ----- 
            if (areaNr1 == 0) {
                // If entire section above water
                crossy1 = this.localyCoord[secNr][areaNr1];
                crossz1 = this.localzCoord[secNr][areaNr1];
                this.firstWetLengthStarb[secNr] = 0;
            } else {
                double x = x_cross1-this.globalStarBxCoord[secNr][areaNr1-1];
                double y = y_cross1-this.globalStarByCoord[secNr][areaNr1-1];
                double z = z_cross1-this.globalStarBzCoord[secNr][areaNr1-1];
                double wetLength = Math.sqrt(x*x + y*y + z*z);
                this.firstWetLengthStarb[secNr] = wetLength;
                
                crossy1 = this.localyCoord[secNr][areaNr1-1]
                        +(wetLength/this.elementLength[secNr][areaNr1-1]) 
                        *(this.localyCoord[secNr][areaNr1]-this.localyCoord[secNr][areaNr1-1]);
                crossz1 = this.localzCoord[secNr][areaNr1-1]
                        +(wetLength/this.elementLength[secNr][areaNr1-1]) 
                        *(this.localzCoord[secNr][areaNr1]-this.localzCoord[secNr][areaNr1-1]);
            }
            
            // ----- Aport side -----  
            if (areaNr2 == 0) {
                // If entire section above water
                crossy2 = -this.localyCoord[secNr][areaNr2];
                crossz2 = this.localzCoord[secNr][areaNr2];
                this.firstWetLengthAPort[secNr] = 0;
            } else {
                double x = x_cross2-this.globalAPortxCoord[secNr][areaNr2-1];
                double y = y_cross2-this.globalAPortyCoord[secNr][areaNr2-1];
                double z = z_cross2-this.globalAPortzCoord[secNr][areaNr2-1];
                double wetLength = Math.sqrt(x*x + y*y + z*z);
                this.firstWetLengthAPort[secNr] = wetLength;
                
                crossy2 = -this.localyCoord[secNr][areaNr2-1]
                        -(wetLength/this.elementLength[secNr][areaNr2-1]) 
                        *(this.localyCoord[secNr][areaNr2]-this.localyCoord[secNr][areaNr2-1]);
                crossz2 = this.localzCoord[secNr][areaNr2-1]
                        +(wetLength/this.elementLength[secNr][areaNr2-1]) 
                        *(this.localzCoord[secNr][areaNr2]-this.localzCoord[secNr][areaNr2-1]);
            }
            
            this.crossyCoordLocalStarB[secNr] = crossy1;
            this.crosszCoordLocalStarB[secNr] = crossz1;
            this.crossyCoordLocalAPort[secNr] = crossy2;
            this.crosszCoordLocalAPort[secNr] = crossz2;
        }
    }
}
