package no.marintek.simulator.ses;
import no.marintek.simulator.generic.GenericFederate;
import no.marintek.simulator.sensor.MRUData;


/**
 * Communicates with LouverFederate and SESFederate.
 * 
 * Receives cg_down from vessel (correct vertical position in the NED frame)
 * Receives acc_heave_MRU from the MRU federation.
 * This is integrated up to pos_heave_MRU which corresponds to cg_down but
 * the sensor has some drifting error. Must be fixed!
 * 
 * Receives the max available leakage area from the LouverFederate (A_L_Louver_Max.
 * The Heave Controller calculates a leak area between 0 and this max value.
 * This leak will damp vertical movement (P-Controller).
 * 
 * This leakage area (A_L_Active_ctr) is transformed into a corresponding louver 
 * handle position for the LouverFederate (handle_pos_ctrl, 0-100, 0 i closed)
 * 
 * This value is sent to the LouverFederate that will put its handle to the wanted 
 * position.
 * 
 * controlled_airflow denotes percentage of possible accessable actuator airflow.
 * The value will differ from 0 to 100 based on desired actuator behavior
 * if 0: louver / fan airflow is zero
 * if 100: louver / fan airflow is maxed based on the specified physical actuators contribution
 *  
 * 
 * @author Øyvind Auestad
 */

public class HeaveControl extends GenericFederate {
    
    double noisyPosMeasurement;

    private boolean louver_control_active = false;
    private boolean fan_control_active = false;
    private boolean update_air_cushion = true;
    private boolean ref_point_calibrated = false;
    // For auto detection of reference position:

    private double time, dt;
    private double pos_heave_REF = -1.8; // Only some init. condition, will be calibrated
    
    // Max physical leak area from louvers (specified by LouverFederate CSI parameter)
    private double A_L_Louver_Max = 0;
    
    // Q_in from fan characteristic. It is the maximum air inflow for control
    private double Q_in_characteristic = 0;
    
    // See the introduction at the top for explanation of controlled_airflow_louver [%]
    private int controlled_louver_pos; //(0-100)
    private int controlled_fan_pos; // (0-100)
    
    // Proportional gain for P-controller. Fetched from CSI
    //private double K_p_fan, K_p_louver;
    //private double K_d_fan, K_d_louver;
    private double K_p, K_d;
    
    // Down position of the NED system. The prescript explains where on the boat
    // and down is the current point of control (only 1 of them at the time)
    private double cg_down, bow_down, down, prev_down, mru_down;
    
    private String cg_down_OR_bow_down;
    
    private double cg_heave_vel;
               
    // Height position from MRU (MRU is for now placed at cg)
    // Error drifting occur. Must be fixed.
    // Chooese wheater you wanna use cg_down or pos_heave_MRU as input for Heave Controller
    private double acc_heave_MRU = 0; 
    private double vel_heave_MRU = 0;
    private double pos_heave_MRU = 0;
    
    // Decides what to control. CG (if 0) or vessel bow (if 1)
    private int point_of_control;
    private int point_of_control_prev;
    
    // Tuning kalmanfilter and LQR controller (LQG controll)
    double stdv_x, stdv_z, Q_diag, R_diag;
    
    // For P-Controller:
    private double error;
    private double error_dot;
    private double u;
    private double u_louver;
    private double u_fan;
    
    private MRUData mru;
    private KalmanFilter kalman;
    
    // Change the actuator resolution (20 steps from closed to open actuator)
    private int resolutionFan = 100;
    private int resolutionLouver = 100;
            
    public HeaveControl(String configFile, String simType){
        super(configFile, simType);
        logger.info("Starting simulation");
        System.out.println("config1");
        simulate();
    }
    
    public HeaveControl(String configFile, String fedName, String simType){
        super(configFile, fedName, simType);
        logger.info("Starting simulation");
        System.out.println("config2");
        simulate();
    }
    
    private void simulate(){
        simApp.startSimulation();
        System.out.println("Heave controller is calibrating reference point in eq. state. Please wait..");
        point_of_control_prev = simApp.getOwnParameter("point_of_control","Case1_1.HeaveControl").getValueAsInteger();
        mru = new MRUData(simApp);
        while (simApp.isConnected()){
            performStep();
        }
        
        simApp.stopSimulation();
    }
    
    private void performStep(){ 

        // Read attributes
        time = simApp.getReceivedObjectAttribute("time", "SimulationTime").getValueAsDouble();
        this.Q_in_characteristic = simApp.getReceivedObjectAttribute("Q_in_characteristic", "Case1_1.Fan").getValueAsDouble();
        
        // Tuning Kalman / LQR controller:
        // Get parameters:
        stdv_x = simApp.getOwnParameter("stdv_x").getValueAsDouble();
        stdv_z = simApp.getOwnParameter("stdv_z").getValueAsDouble();
        
        dt = simApp.getTimeStep();
             
        K_p = simApp.getOwnParameter("K_p", "Case1_1.HeaveControl").getValueAsDouble();
        K_d = simApp.getOwnParameter("K_d", "Case1_1.HeaveControl").getValueAsDouble();
        
        if (update_air_cushion){
            this.updateHeaveControlSystem();
        }

        down = simApp.getReceivedObjectAttribute(cg_down_OR_bow_down, "Case1_1.AirCushion").getValueAsDouble();
        
        if (!ref_point_calibrated && time > 11){
            // Wait until the vessel has settled:
            // Requires startup in 0 waves and 10 sec before the vessel has settled down
            // Then use this pos as reference heave position.
            System.out.println("Calibrating reference control point...");
            if (Math.round(100000*down) == Math.round(100000*prev_down)){ // Tidligere dette / 2
                System.out.println(100000*down +" - "+100000*prev_down);
                ref_point_calibrated = true;
                simApp.getOwnParameter("ref_point_calibrated").setValueAsBoolean(ref_point_calibrated);
                pos_heave_REF = down;
                simApp.getOwnParameter("pos_heave_REF").setValueAsDouble(pos_heave_REF);
               // System.out.println("MRU = "+(mru_down - pos_heave_REF) + " cg_down = "+down);
                System.out.println("Calibration succesful.. Reference point calibrated to: "+pos_heave_REF);
                
                pos_heave_MRU = pos_heave_REF;
                updateHeaveControlSystem();
                
                // The federate class AirCushion needs to update after this:
                //update_air_cushion
                //simApp.getOwnAttribute("update_air_cushion").setValueAsBoolean(true);
            }
        }
 
        if(ref_point_calibrated){

           if(Math.random() > 0.5){noisyPosMeasurement = down + 0.12*Math.random();}
           else{noisyPosMeasurement = down - 0.12*Math.random();}
           // noisyPosMeasurement = down; // TODO
            
            kalman.getSensorReadings(noisyPosMeasurement);
            kalman.offlineParameterCalibration(stdv_x, stdv_z, K_p, K_p);
            kalman.onlineStateEstimation(1, A_L_Louver_Max, Q_in_characteristic, u);
            simApp.getOwnAttribute("x_est_pos","Case1_1.HeaveControl").setValueAsDouble(kalman.getEstimatedPos());
            simApp.getOwnAttribute("x_est_vel","Case1_1.HeaveControl").setValueAsDouble(kalman.getEstimatedVel());
            
            
            // useLQGControl();
            usePController();
                    
                    
            // See the introduction at the top for explanation of controlled_airflow_louver
           // In short terms: controlled_airflow is percentage of possible accessable actuator air flow
            if(louver_control_active){
                // Louver control
                controlled_louver_pos = getLouverPos();

            }else{
                controlled_louver_pos = 50;}
            
            //Fan control
            if(fan_control_active){
//               controlled_fan_pos = getFanPos();
               if (controlled_louver_pos < 50){
                   controlled_fan_pos = 100;
               }else{ 
                   controlled_fan_pos = 100 - 2*(controlled_louver_pos-50);
               }

            }else{
                controlled_fan_pos = 100;} 
        
        }else{
            controlled_fan_pos = 100; // TODO buggen
            controlled_louver_pos = 50; 
        }

        prev_down = down;
        update_air_cushion = simApp.getReceivedObjectAttribute("update_air_cushion", "Case1_1.AirCushion").getValueAsBoolean();
        
        // Update attributes that are beeing sent to LouverFederate:
        simApp.getOwnAttribute("controlled_louver_pos","Case1_1.HeaveControl").setValueAsInteger(controlled_louver_pos);
        simApp.getOwnAttribute("controlled_fan_pos","Case1_1.HeaveControl").setValueAsInteger(controlled_fan_pos);
        
        // Kalman attributes
        simApp.getOwnAttribute("noisyPosMeasurement","Case1_1.HeaveControl").setValueAsDouble(noisyPosMeasurement);
        simApp.getOwnAttribute("u","Case1_1.HeaveControl").setValueAsDouble(u);
        
        
        simApp.performStep();
    }
    
    private void updateHeaveControlSystem(){
        
        // Control cg or bow: point_of_control respectively 0 and 1.
        this.point_of_control = simApp.getOwnParameter("point_of_control","Case1_1.HeaveControl").getValueAsInteger();
        
        // Choose whether to control cg or the bow of the vessel
        if (point_of_control == 0){
            cg_down_OR_bow_down = "cg_down";
            // Set control parameters to default values

            K_p = 1.5;
            K_d = 0.3;
        }
        else if (point_of_control == 1){
            cg_down_OR_bow_down = "bow_down";
            K_p = 1;
            K_d = 0.2;
        }
        else{ System.out.println("point_of_control must be set to either 0 (control of cg) or 1 (control of the bow)");}
        
        // Read parameters:
        this.louver_control_active = simApp.getOwnParameter("louver_control_active").getValueAsBoolean();
        this.fan_control_active = simApp.getOwnParameter("fan_control_active").getValueAsBoolean();
        this.ref_point_calibrated = simApp.getOwnParameter("ref_point_calibrated","Case1_1.HeaveControl").getValueAsBoolean();

        
        simApp.getOwnParameter("K_p", "Case1_1.HeaveControl").setValueAsDouble(K_p);
        simApp.getOwnParameter("K_d", "Case1_1.HeaveControl").setValueAsDouble(K_d);
        
        double cg_and_bow_height_diff = simApp.getOwnParameter("cg_and_bow_height_diff","Case1_1.HeaveControl").getValueAsDouble();

        this.A_L_Louver_Max = simApp.getReceivedObjectAttribute("A_L_Louver_Max", "Case1_1.Louver").getValueAsDouble();  
        this.Q_in_characteristic = simApp.getReceivedObjectAttribute("Q_in_characteristic", "Case1_1.Fan").getValueAsDouble();
        this.pos_heave_REF = simApp.getOwnParameter("pos_heave_REF").getValueAsDouble();

        // If one changes between cg and bow control while NOT doing a re-calibration 
        // of the reference point (requires 0 waves), this will automaticly 
        // add or subtract the height difference
        pos_heave_REF += (point_of_control_prev - point_of_control)*cg_and_bow_height_diff;
        simApp.getOwnParameter("pos_heave_REF").setValueAsDouble(pos_heave_REF);
        point_of_control_prev = point_of_control;

        kalman = new KalmanFilter(dt, pos_heave_REF);
        
    }
    
    private void usePController(){
        
        
        //error_dot = 0-kalman.getEstimatedVel();
        //u = (error + error_dot);

        
        error = pos_heave_REF - kalman.getEstimatedPos();
        u = error;
        double u_dot = - kalman.getEstimatedVel();
        
        // Saturate
        double sat_louver = A_L_Louver_Max / (2*K_p);
        double sat_fan = Q_in_characteristic /(K_p);
        if (u > sat_louver){
            u = sat_louver;
        }else if(u < -sat_louver){
            u = -sat_louver;
        }
        
        // PD
        this.u_louver = K_p * u + K_d * u_dot;
       // this.u_fan = - K_p_fan * u - K_d_fan * u;
        //System.out.println(K_d_fan * u + " - "+K_d_louver * u+ " - "+ K_p_fan * u+" - "+ K_p_louver * u);
  
    }
    
     public int getLouverPos(){

         // A scaled value between 0 and 100. 0: no louver leakage, 100: max louver leak area
        
        // Saturation. The leak cant be bigger than the total louver area specified
        // in LouverFederate. Leakage area also is > 0. The saturation is:
        // A_L_Louver_Max >= u_louver + A_L_Louver_Max/2 >= 0
         if (u_louver <= -A_L_Louver_Max/2){
            u_louver = - A_L_Louver_Max/2;
        }
        else if(u_louver >= A_L_Louver_Max/2){
            u_louver = A_L_Louver_Max/2;
        }
        
        // Wanted area for the louvers in order to control heave movement
        // The louver area leakage that is used for cushion calculataion (in SESFederate)
        // will have a actuator delay which is implemented in FanFederate
        double A_L_Louver_ctrl = u_louver + A_L_Louver_Max/2;
        // Louver openess in percentage. If 0: louver is closed, if 100: its 100% open
        int controlled_louver_pos = (int)(100 * (A_L_Louver_ctrl/A_L_Louver_Max));
        // Round off to actuator resolution (example: 20 steps from closed to open)
        // With 20 steps: Function will return a number from 0 to 100 rounded iff to closest 5.
        return (int)(Math.round(controlled_louver_pos *resolutionLouver/100.0) * (100.0/resolutionLouver));
    }
    
    public int getFanPos(){
        // A scaled integer from 0 and 100. 0: no fan contribution, 100: max contr.
        // Saturation: Q_in_char >= u + Q_in_char/2 >= 0
        
        
        if (u_fan <= -Q_in_characteristic){ // delt på 2 her
            u_fan = -Q_in_characteristic;}
        else if(u_fan > 0){
             u_fan = 0;}
        
        // Air flow from fans which is the controll value. The desired inflow from the fans.
        double Q_in_ctrl = u_fan + Q_in_characteristic; // og denne
        
        int controlled_fan_pos = (int)(100 * (Q_in_ctrl/Q_in_characteristic));
        
        // Rounds off controlled_fan_pos to closest (100/resolutionFan)
        return (int)(Math.round(controlled_fan_pos *resolutionFan/100.0) * (100.0/resolutionFan));
    }
    
    public static void main(String[] args){
        // This class needs input from the command line to start correctly
        String federate = null;         // First input argument
        String simBus = null;           // Second input argument
        String federationFile = null;   // Third input argument
               
        if(args.length > 1){
            federationFile = args[0];
            if(args[1].equalsIgnoreCase("csi") || args[1].equalsIgnoreCase("hla")){
                simBus = args[1];
                System.out.println(simBus);
                HeaveControl heavecontroller = new HeaveControl(federationFile, simBus);
            } else {
                federate = args[1];
            }
        } 
        else{
            System.out.println("Federation file and sim bus type must be specified");
        }
        
        if(args.length > 2){
            federate = args[1];
            simBus = args[2];
            HeaveControl updateHeaveControlSystem = new HeaveControl(federationFile,federate, simBus);
        }
    }       
}