package no.marintek.simulator.ses;
import no.marintek.simulator.generic.GenericFederate;

/**
 * Louver
 * 
 * Input: handle_pos_ctrl
 * the input is what the handle position should be according to the Ride controller.
 * 
 * stepDelay differs this position from the real handle position (handle_position)
 * 
 * Output: A_L_Active: 
 * The area leak that is produced due to what handle_position is valued
 * 
 * @author Øyvind Auestad
 */

public class Louver extends GenericFederate {
    
    private double time = 0;
    private boolean update_air_cushion = true;
    
    private double A_L_Louver_Max= 0; //[m]
    private int number_of_louvers = 0;
    
    // Indicates in percentage the opening of the louvers:
    private int controlled_louver_pos = 0; // Louver pos ordered by Heave controller (0 -100)
    private int louver_pos = 0; // actual louver_position (0 - 100)
    // Because of actuator delay louver_pos is always slower than what the controller 
    // wants.
    
    private double controlled_louver_pos_arrival_time = 0; //[s]
    // stepDelay: Specifies the step delay. 
    // Thus: stepDelay = 0.01 -> Moving louver handle from pos 0 to 1 takes 0.01sek
    // and move louver handle from pos 0 to 20 takes 20*0.01sek = 0.20 sek
    private double open_close_actuator_delay = 0;
    
    private double A_L_Louvers = 0; // Current leakage area from louvers
    double A_L_Mean_Louvers = 0;
    
    // The gap between current handle position (0-100) and the next handle 
    // position ordered by the Heave controller.
    private int gap = 0;

    public Louver(String configFile, String simType){
        super(configFile, simType);
        logger.info("Starting simulation");
        simulate();
    }
    
    public Louver(String configFile, String fedName, String simType){
        super(configFile, fedName, simType);
        logger.info("Starting simulation");
        simulate();
    }
    
    private void simulate(){
        simApp.startSimulation();
        while (simApp.isConnected()){
            performStep();
        }
        simApp.stopSimulation();
    }
    
    private void performStep(){ 
        // Get received objects:
        update_air_cushion  = simApp.getReceivedObjectAttribute("update_air_cushion", "Case1_1.AirCushion").getValueAsBoolean(); 
        time = simApp.getReceivedObjectAttribute("time", "SimulationTime").getValueAsDouble();
        controlled_louver_pos = simApp.getReceivedObjectAttribute("controlled_louver_pos", "Case1_1.HeaveControl").getValueAsInteger();
         
        open_close_actuator_delay  = simApp.getOwnParameter("open_close_actuator_delay", "Case1_1.Louver").getValueAsDouble(); // Delay: sec/% 
        A_L_Louver_Max = simApp.getOwnParameter("A_L_Louver_Max", "Case1_1.Louver").getValueAsDouble();
        number_of_louvers = simApp.getOwnParameter("number_of_louvers", "Case1_1.Louver").getValueAsInteger();


        if (actuatorInWantedPosition(open_close_actuator_delay, time)){
            this.louver_pos = controlled_louver_pos;
        }
        // Rewriting: louver_pos = 100*(A_L_Louvers)/A_L_Louver_Max
        A_L_Louvers = (A_L_Louver_Max/100) * (louver_pos);
        
        // Mean louver leak is defined as 50% of maximum leak:
        A_L_Mean_Louvers = (A_L_Louver_Max/2);
        
        // For comparing controlled and actual louver handle position:
        // To see this clearly one can set resolutionLouver (in RideController.java) to 1
       // System.out.println("Louver position wanted by controller = "+controlled_louver_pos + "--- Actual louver position = "+louver_pos);
                
        simApp.getOwnAttribute("A_L_Louvers", "Case1_1.Louver").setValueAsDouble(A_L_Louvers * number_of_louvers);
        simApp.getOwnAttribute("A_L_Mean_Louvers", "Case1_1.Louver").setValueAsDouble(A_L_Mean_Louvers * number_of_louvers); 
        
        //Update own attribute so that it can be sent over the CSI bus (only attributes can be sent)
        simApp.getOwnAttribute("A_L_Louver_Max", "Case1_1.Louver").setValueAsDouble(A_L_Louver_Max );
        
        // For viualization only:
        simApp.getOwnAttribute("louver_pos", "Case1_1.Louver").setValueAsDouble(louver_pos);
     
        simApp.performStep();   
    }
    
    private boolean actuatorInWantedPosition(double open_close_actuator_delay, double time){
        // Tells if handle is in the current wanted pos (from Heave)
        if(gap*(open_close_actuator_delay/100) + controlled_louver_pos_arrival_time <= time ){
            // In position now!
            controlled_louver_pos_arrival_time = time;
            gap = Math.abs(louver_pos - controlled_louver_pos);
            return true;
        }else{
            // Not in position!
            return false;
        }
    }
    
    public static void main(String[] args){
        // This class needs input from the command line to start correctly
        String federate = null;         // First input argument
        String simBus = null;           // Second input argument
        String federationFile = null;   // Third input argument
               
        if(args.length > 1){
            federationFile = args[0];
            if(args[1].equalsIgnoreCase("csi") || args[1].equalsIgnoreCase("hla")){
                simBus = args[1];
                System.out.println(simBus);
                Louver louver = new Louver(federationFile, simBus);
            } else {
                federate = args[1];
            }
        } 
        else{
            System.out.println("Federation file and sim bus type must be specified");
        }
        
        if(args.length > 2){
            federate = args[1];
            simBus = args[2];
            Louver louver = new Louver(federationFile,federate, simBus);
        }
    }
}
