/*
 * SESForces
 * Calculates air cushion forces due to overpressure in air cushion
 * Positive directions of forces follow the air cushion axis system.
 * It also calculates the viscous forces acting on the vessel.
 * 
 * Made by Halvorsen/Auestad, Spring 2008/2011
 */

/* 
Calculate forces:
Forces are calculated in the air cushion axis system.
Relative difference to axis system intended for forces
Forces stored in force_* etc is in accordance with the vesim input system (positive directions)

Forces due to air cushion overpressure:
Forces are given as relative static forces if on cushion (initData)
Forces are negative static forces if off cushion 
(the vessel federate is assuming that mass balance is fullfilled at given static draught. 
This is only the case if on cushion. If off cushion the static lift is lost)*/

package no.marintek.simulator.ses;

import no.marintek.csi.datadesk.IObjectStorage;
import no.marintek.tools.dataobjects.Ned;
import no.marintek.tools.math.CoordinateConversions;
import no.marintek.tools.dataobjects.Motion;
import no.marintek.simulator.vessel.ShipMotionConversion;
import no.marintek.tools.dataobjects.Linear;

public class SESForces {

    // For cushion forces:
    private double cushion_force_surge = 0;
    private double cushion_force_sway = 0;
    private double cushion_force_heave = 0;
    private double cushion_moment_roll = 0;
    private double cushion_moment_pitch = 0;
    private double cushion_moment_yaw = 0;
    
    public double cforce_surge = 0;
    public double cforce_sway = 0;
    public double cforce_heave = 0;
    public double cmoment_roll = 0;
    public double cmoment_pitch = 0;
    public double cmoment_yaw = 0;
    
    
    private double added_viscous_surge = 0;
    private double added_viscous_pitch = 0;
    private double calm_viscous_surge, calm_viscous_pitch = 0;
    private double total_viscous_pitch, total_viscous_surge = 0;
    
    private AirCushionGeometry cushion;
    private double p_c;
    
    // For viscous forces:
    private IObjectStorage object;
    private Ned ned_wave_vel;
    private SESHull hull;
    private double draught, cg_surge_vel, cur_surge_vel, wetSurface;
    private Constants constant;
    private ShipMotionConversion convertShipMotion;
    private Motion cgMotion;

    private InitialSESData initData;
      
    public SESForces(AirCushionGeometry cushion, double p_c){
        
        // For calculating cushion forces
        this.cushion = cushion;
        this.p_c = p_c;
        
        calculateCushionForces();            
    }
    
    private void calculateCushionForces() {
        
            cforce_surge = 0;
            cforce_sway = 0;
            cforce_heave = 0;
            cmoment_roll = 0;
            cmoment_pitch = 0;
            cmoment_yaw = 0;
       
        for (int secNr=0; secNr<cushion.sectionNumber; secNr++) {
            double dx = cushion.dx[secNr];
            double dsway = 0;
            double dheave = 0;
            
            // ----- Starboard side ----- 
            // Forces between crossing point of water and hull and the first dry coordinate node
            dsway = p_c * (cushion.localzCoord[secNr][cushion.firstCoordInAirStarb[secNr]] 
                    - cushion.crosszCoordLocalStarB[secNr]) * dx; // Z direction
            // Z direction is defined positive downwards
            dheave = p_c * (cushion.crossyCoordLocalStarB[secNr] 
                    - cushion.localyCoord[secNr][cushion.firstCoordInAirStarb[secNr]]) * dx;
            
            cforce_sway += dsway;
            cforce_heave += dheave;
            
            cmoment_roll += dsway*0.5*(cushion.localzCoord[secNr][cushion.firstCoordInAirStarb[secNr]] 
                    + cushion.crosszCoordLocalStarB[secNr])
                    - dheave*0.5*(cushion.localyCoord[secNr][cushion.firstCoordInAirStarb[secNr]] 
                    + cushion.crossyCoordLocalStarB[secNr]);
            cmoment_pitch += dheave*cushion.MomentArmx[secNr];
            cmoment_yaw += -dsway*cushion.MomentArmx[secNr];
            
            // Forces on elements that are completely dry
            for (int elemNr=cushion.firstCoordInAirStarb[secNr]; elemNr<cushion.coordNumber[secNr]-1; elemNr++) {
                dsway = -p_c*cushion.elementLength[secNr][elemNr]*cushion.yNormalVector[secNr][elemNr]*dx;
                dheave = -p_c*cushion.elementLength[secNr][elemNr]*cushion.zNormalVector[secNr][elemNr]*dx;
                cforce_sway += dsway;
                cforce_heave += dheave;
                cmoment_roll += dsway*cushion.MomentArmz[secNr][elemNr] - dheave*cushion.MomentArmy[secNr][elemNr];
                cmoment_pitch += dheave*cushion.MomentArmx[secNr];
                cmoment_yaw += -dsway*cushion.MomentArmx[secNr];
            }
            
            // ----- Aport side -----
            // Forces between crossing point of water and hull and the first dry coordinate node
            dsway = -p_c * (cushion.localzCoord[secNr][cushion.firstCoordInAirAPort[secNr]] 
                    - cushion.crosszCoordLocalAPort[secNr]) * dx;
            dheave = p_c * (-cushion.localyCoord[secNr][cushion.firstCoordInAirAPort[secNr]]
                    -cushion.crossyCoordLocalAPort[secNr]) * dx;

            cforce_sway += dsway;
            cforce_heave += dheave;
            cmoment_roll += dsway*0.5*(cushion.localzCoord[secNr][cushion.firstCoordInAirAPort[secNr]] 
                    + cushion.crosszCoordLocalAPort[secNr])
                    - dheave*0.5*(-cushion.localyCoord[secNr][cushion.firstCoordInAirAPort[secNr]] 
                    + cushion.crossyCoordLocalAPort[secNr]);
            cmoment_pitch += dheave*cushion.MomentArmx[secNr];
            cmoment_yaw += -dsway*cushion.MomentArmx[secNr];
            
            // Forces on elements that are completely dry
            // The vectors and moment arms are meant for starboard side, 
            // some of them must therefore be multiplied by -1 to obtain correct direction
            for (int elemNr=cushion.firstCoordInAirAPort[secNr]; elemNr<cushion.coordNumber[secNr]-1; elemNr++) {
                dsway = p_c*cushion.elementLength[secNr][elemNr]*cushion.yNormalVector[secNr][elemNr]*dx;
                dheave = -p_c*cushion.elementLength[secNr][elemNr]*cushion.zNormalVector[secNr][elemNr]*dx;
                cforce_sway += dsway;
                cforce_heave += dheave;
                cmoment_roll += dsway*cushion.MomentArmz[secNr][elemNr] + dheave*cushion.MomentArmy[secNr][elemNr];
                cmoment_pitch += +dheave*cushion.MomentArmx[secNr];
                cmoment_yaw += -dsway*cushion.MomentArmx[secNr];
            }
        } 
        
    }
    
    private void calculateViscousForces(){
        
        // Calculates viscous forces

        // Convert wave particle velocities to body system
        CoordinateConversions cv = new CoordinateConversions();
      
        Motion vesselMotion = convertShipMotion.getMotionAtPoint(hull.Lpp/2, 0.0, draught/2, cgMotion);  // x,y,z-pos of wave particle where finding velocity
        Linear body_wave_vel = cv.transformLinearVelocityFromNEDToBody(vesselMotion.angularDisplacement,ned_wave_vel); // Find velocities in local body system

        double V = cg_surge_vel - cur_surge_vel - body_wave_vel.surge;                  // All velocities are in local direction
        double Rn = Math.abs(V)*hull.Lpp/constant.v;                                    // Reynold's number
        double CF = 0.075/Math.pow((Math.log10(Rn)-2),2);                               // ITTC-57 frictional line
        double deltaCF = (110.31*Math.pow(constant.H*Math.abs(V), 0.21)-403.33)*CF*CF;  // Correction for hull roughness
        deltaCF = Math.max(deltaCF, 0);                                                 // Correction for hull roughness must be 0 or larger
        double CV = (1+constant.k)*(CF+deltaCF);                                        // Viscous resistance coefficient

        // Calculate viscous forces
        this.total_viscous_surge = -0.5*constant.rho*V*Math.abs(V)*CV;//*(this.wetSurface)*CV;                                
        // Total viscous resistance gives this pitch moment about center of air cushion axis system
        this.total_viscous_pitch = -total_viscous_surge*draught/2;                                                    
        // Viscous resistance calm water
        this.calm_viscous_surge = -0.5*constant.rho*V*Math.abs(V)*initData.getwet_surface_0()*CV;      
        // Calm water viscous resistance gives this pitch moment about center of air cushion axis system
        this.calm_viscous_pitch = -calm_viscous_surge*draught/2;                                                      
        
        this.added_viscous_surge = -0.5*constant.rho*V*Math.abs(V)*(this.wetSurface-initData.getwet_surface_0())*CV;  // Added viscous resistance
        this.added_viscous_pitch = -added_viscous_surge*draught/2;                                                    // Added viscous resistance gives this pitch moment about center of air cushion axis system
        
    }
    
    public void updateForces(boolean on_cushion, boolean viscous_forces_active, 
            InitialSESData initData, double p_c, AirCushionGeometry cushion, 
            ShipMotionConversion convertShipMotion, Ned ned_wave_vel, 
            SESHull hull, double draught, double cg_surge_vel, double cur_surge_vel,
            Constants constant, double wetSurface, Motion cgMotion){
        
        // For calculating cushion forces
        this.cushion = cushion;
        this.p_c = p_c;
        this.initData = initData;
        
        // For calculating viscous forces:
        this.convertShipMotion = convertShipMotion;
        this.ned_wave_vel = ned_wave_vel;
        this.hull = hull;
        this.draught = draught;
        this.cg_surge_vel = cg_surge_vel;
        this.cur_surge_vel = cur_surge_vel;
        this.constant = constant;
        this.wetSurface = wetSurface;
        this.cgMotion = cgMotion;   
        
        // Add the static contribution (with p_0)
        cushion_force_surge = -initData.force_surge;
        cushion_force_sway = -initData.force_sway;
        cushion_force_heave = initData.force_heave;

        cushion_moment_roll = -initData.moment_roll;
        cushion_moment_pitch = -initData.moment_pitch;
        cushion_moment_yaw = initData.moment_yaw;

        if (on_cushion){
            
            calculateCushionForces();
           
            cushion_force_surge += cforce_surge;
            cushion_force_sway += cforce_sway;
            cushion_force_heave += -cforce_heave;


            cushion_moment_roll += cmoment_roll;
            cushion_moment_pitch += cmoment_pitch;
            cushion_moment_yaw += -cmoment_yaw;
        }
        if (viscous_forces_active){
            calculateViscousForces();
        }
    }
    
    // Get cushion forces
    public double getCushionSurge() { return cushion_force_surge; }
    public double getCushionSway() { return cushion_force_sway; }    
    public double getCushionHeave() { return cushion_force_heave; }    
    public double getCushionRoll() { return cushion_moment_roll; }    
    public double getCushionPitch() { return cushion_moment_pitch; }    
    public double getCushionYaw() { return cushion_moment_yaw; }
    
    // Get viscous forces
    public double getAddedViscousSurge(){return added_viscous_surge;}
    public double getAddedViscousPitch(){return added_viscous_pitch;}

    // For monitoring values only (has not a direct effect on anything 
    // except calculating the two atttributes stated above
    
    public double getCalmViscousSurge(){return calm_viscous_surge;}
    public double getCalmViscousPitch(){return calm_viscous_pitch;}
    public double getTotalViscousSurge(){return total_viscous_surge;}
    public double getTotalViscousPitch(){return total_viscous_pitch;}
    
    
    

    
}