/*
 * hull
 * Contains geometry of the vessel
 * 
 * This federate uses the mgf axis system
 * It is used to calculate wet surface, needed to calculate viscous resistance
 * 
 * Made by Trygve Espeland Halvorsen, Spring 2008
 */

package no.marintek.simulator.ses;

import no.marintek.tools.dataobjects.Motion;
import no.marintek.simulator.vessel.ShipMotionConversion;
import no.marintek.simulator.environment.WavesModel;
import java.io.*;
import java.util.*;


public class SESHull {
    int minSections = 40;       // Minimum number of nodes in sections
    
    // Common data
    public double Lpp;          // vessel length
    public int sectionNumber;   // numbers of sections defining the hull
    public int[] coordNumber;   // Numbering following the 0, 1, 2, .. format. Numbered from AP and forward
    public int[] lowestCoordNr; // The node number on the section that is lowest
    
    // Local data
    public double[] localxCoord;
    public double[][] localyCoord;
    public double[][] localzCoord;
    
    // Global data
    public double[][] globalStarBxCoord;
    public double[][] globalStarByCoord;
    public double[][] globalStarBzCoord;  
    public double[][] globalAPortxCoord;
    public double[][] globalAPortyCoord;
    public double[][] globalAPortzCoord; 
    public double[] globaldL;
    
    // Information about where the hull crosses the water surface. Outside the air cushion
    public int[] firstWetCoordStarb;
    public double[] firstWetLengthStarb;
    public int[] firstWetCoordAPort;
    public double[] firstWetLengthAPort;
    
    // Element data
    public double[] dx;
    public double[][] elementLength;
    
    
    // Initialize object, define geometry
    public SESHull(String filePath) {
        this.readmgfFile(filePath);
        this.defineElementLengths();
        
        
        
        
    }
    
    public void updateStatic(Constants constant, double draught) {
        this.findSubmergenceStatic(constant, draught);
    }
    
    public void updateGlobal(Motion cgMotion, ShipMotionConversion convertShipMotion, double time) {
        this.makeGlobalCoord(cgMotion, convertShipMotion);
        this.findSubmergence(time);
    }
    
    
    private void findSubmergence(double time) {
        WavesModel waves = new WavesModel();

        for (int secNr=0; secNr<this.sectionNumber; secNr++){

            // ----- Starboard side ----- 
            double x_cross1;
            double y_cross1;
            double z_cross1;
            double x_cross2;
            double y_cross2;
            double z_cross2;
            int wetCoordNr1;
            int wetCoordNr2;
            
            boolean goon = true;
            double submergence = 0; // Positive means above water
            double previousSubmergence = 0;
            double wave_elevation_xy;
            int candidate = -1;
            
            while (goon) {  
                candidate++;
                previousSubmergence = submergence;
                wave_elevation_xy = waves.getInterpolatedWaveElevation(time, 
                        this.globalStarBxCoord[secNr][candidate], this.globalStarByCoord[secNr][candidate]);   // pos. down
                submergence = -this.globalStarBzCoord[secNr][candidate]+wave_elevation_xy;  // Positive means point above water
                if (submergence < 0)    // If the point is below water - stop loop                   
                    goon = false;
                else if (candidate == this.coordNumber[secNr]-1)
                    goon = false;            
            } 
            wetCoordNr1 = candidate;
            
            // If entire section submerged or above water
            if (candidate == 0
                    || ((candidate==(this.coordNumber[secNr]-1)) && (submergence > 0))){
                x_cross1 = this.globalStarBxCoord[secNr][candidate];
                y_cross1 = this.globalStarByCoord[secNr][candidate];
                z_cross1 = this.globalStarBzCoord[secNr][candidate];
            } else {
                // Interpolate to find point where water surface crosses hull
                x_cross1 = this.globalStarBxCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalStarBxCoord[secNr][candidate]-this.globalStarBxCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
                y_cross1 = this.globalStarByCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalStarByCoord[secNr][candidate]-this.globalStarByCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
                z_cross1 = this.globalStarBzCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalStarBzCoord[secNr][candidate]-this.globalStarBzCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
            }
            
            // ----- Aport side ----- 
            goon = true;
            submergence = 0;    // Positive means above water
            previousSubmergence = 0;
            candidate = -1;
            while (goon) {  
                candidate++;
                previousSubmergence = submergence;
                wave_elevation_xy = waves.getInterpolatedWaveElevation(time, 
                        this.globalAPortxCoord[secNr][candidate], this.globalAPortyCoord[secNr][candidate]);   // pos. down
                submergence = -this.globalAPortzCoord[secNr][candidate]+wave_elevation_xy;  // Positive means point above water
                if (submergence < 0)    // If the point is below water - stop loop                   
                    goon = false;
                else if (candidate == this.coordNumber[secNr]-1)
                    goon = false;            
            } 
            wetCoordNr2 = candidate;

            // If entire section submerged or above water
            if (candidate == 0
                    || ((candidate==(this.coordNumber[secNr]-1)) && (submergence > 0))){
                x_cross2 = this.globalAPortxCoord[secNr][candidate];
                y_cross2 = this.globalAPortyCoord[secNr][candidate];
                z_cross2 = this.globalAPortzCoord[secNr][candidate];
            } else {              
                // Interpolate to find point where water surface crosses hull
                x_cross2 = this.globalAPortxCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalAPortxCoord[secNr][candidate]-this.globalAPortxCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
                y_cross2 = this.globalAPortyCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalAPortyCoord[secNr][candidate]-this.globalAPortyCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
                z_cross2 = this.globalAPortzCoord[secNr][candidate-1] - previousSubmergence
                        * (this.globalAPortzCoord[secNr][candidate]-this.globalAPortzCoord[secNr][candidate-1])
                        / (submergence-previousSubmergence);
            }                                 

            // Store crossing data
            this.firstWetCoordStarb[secNr] = wetCoordNr1;
            this.firstWetCoordAPort[secNr] = wetCoordNr2;
    
            // Find wetted length of first element that touches water

            // ----- Starboard side ----- 
            if (wetCoordNr1 == 0) {
                // If entire section below water
                this.firstWetLengthStarb[secNr] = 0; // Entire element length is wet - is included in the general wet length calculation
            } else {
                double x = x_cross1-this.globalStarBxCoord[secNr][wetCoordNr1];
                double y = y_cross1-this.globalStarByCoord[secNr][wetCoordNr1];
                double z = z_cross1-this.globalStarBzCoord[secNr][wetCoordNr1];
                this.firstWetLengthStarb[secNr] = Math.sqrt(x*x + y*y + z*z);
            }
            
            // ----- Aport side -----  
            if (wetCoordNr2 == 0) {
                // If entire section below water
                this.firstWetLengthAPort[secNr] = 0; // Entire element length is wet - is included in the general wet length calculation
            } else {
                double x = x_cross2-this.globalAPortxCoord[secNr][wetCoordNr2];
                double y = y_cross2-this.globalAPortyCoord[secNr][wetCoordNr2];
                double z = z_cross2-this.globalAPortzCoord[secNr][wetCoordNr2];
                this.firstWetLengthAPort[secNr] = Math.sqrt(x*x + y*y + z*z);
            }
        }
    }
    
    
    private void findSubmergenceStatic(Constants constant, double draught) {
              
        for (int secNr=0; secNr<this.sectionNumber; secNr++){

            // ---------------------------------------------
            // Find the two coordinates on section where hull crosses water surface
            // ---------------------------------------------          
            double y_cross1;
            double z_cross1;
            double y_cross2;
            double z_cross2;
            int wetCoordNr1;
            int wetCoordNr2;
            
            boolean goon = true;
            double submergence = 0; // Positive means above water
            double previousSubmergence = 0;
            int candidate = -1;

            // positive z is up
            while (goon) {     
                candidate++;
                previousSubmergence = submergence;
                submergence = this.localzCoord[secNr][candidate] - draught;
                if (submergence < 0) { // If the point is below water - stop loop                   
                    goon = false; }
                else if (candidate == (this.coordNumber[secNr]-1)) {    // If the point is the last point (at the center line) - stop loop
                    goon = false; }            
            } 
            wetCoordNr1 = candidate;
            
            // If entire section submerged or above water
            if (candidate == 0
                    || ((candidate==(this.coordNumber[secNr]-1)) && (submergence > 0))){
                //System.out.println("Case 1, leakage or totally submerged");
                y_cross1 = this.localyCoord[secNr][candidate];
                z_cross1 = this.localzCoord[secNr][candidate];

            } else {
                // Interpolate to find point where water surface crosses hull
                z_cross1 = this.localzCoord[secNr][candidate] - submergence;
                y_cross1 = this.localyCoord[secNr][candidate-1] +
                        (z_cross1 - this.localzCoord[secNr][candidate-1])
                        * (this.localyCoord[secNr][candidate] - this.localyCoord[secNr][candidate-1])
                        / (this.localzCoord[secNr][candidate] - this.localzCoord[secNr][candidate-1]);
            }

            // Find crossing point on opposite side of CL
            y_cross2 = -y_cross1;
            z_cross2 = z_cross1;
            wetCoordNr2 = wetCoordNr1;               
            
            // Store crossing data
            this.firstWetCoordStarb[secNr] = wetCoordNr1;
            this.firstWetCoordAPort[secNr] = wetCoordNr2;
            
            // Find wetted length of first element that touches water

            // ----- Starboard side ----- 
            if (wetCoordNr1 == 0) {
                // If entire section below water
                this.firstWetLengthStarb[secNr] = 0; // Entire element length is wet - is included in the general wet length calculation
            } else {
                double y = y_cross1-this.localyCoord[secNr][wetCoordNr1];
                double z = z_cross1-this.localzCoord[secNr][wetCoordNr1];
                this.firstWetLengthStarb[secNr] = Math.sqrt(y*y + z*z);
            }
            
            // ----- Aport side -----  
            if (wetCoordNr2 == 0) {
                // If entire section above water
                this.firstWetLengthAPort[secNr] = 0; // Entire element length is wet - is included in the general wet length calculation
            } else {
                double y = y_cross2+this.localyCoord[secNr][wetCoordNr2];
                double z = z_cross2-this.localzCoord[secNr][wetCoordNr2];
                this.firstWetLengthAPort[secNr] = Math.sqrt(y*y + z*z);
            }
        }
    }

    
    private void makeGlobalCoord(Motion cgMotion, ShipMotionConversion convertShipMotion) {
        Motion pointMotion = new Motion();
        
        for (int secNr=0; secNr<this.sectionNumber; secNr++) {

            // ----- Starboard side -----
            for (int coordNr=0; coordNr<this.coordNumber[secNr]; coordNr++) {                     
                pointMotion = convertShipMotion.getMotionAtPoint(
                        this.Lpp/2-this.localxCoord[secNr],
                        this.localyCoord[secNr][coordNr],
                        this.localzCoord[secNr][coordNr], cgMotion);
                this.globalStarBxCoord[secNr][coordNr] = pointMotion.ned.north; // Setting global x-coordinate of section
                this.globalStarByCoord[secNr][coordNr] = pointMotion.ned.east;  // Setting global y-coordinate of section
                this.globalStarBzCoord[secNr][coordNr] = pointMotion.ned.down;  // Setting global z-coordinate of section
            }    
            
            // ----- Aport side -----
            for (int coordNr=0; coordNr<this.coordNumber[secNr]; coordNr++) {  
                pointMotion = convertShipMotion.getMotionAtPoint( 
                        this.Lpp/2-this.localxCoord[secNr],
                        -this.localyCoord[secNr][coordNr],
                        this.localzCoord[secNr][coordNr], cgMotion);
                this.globalAPortxCoord[secNr][coordNr] = pointMotion.ned.north; // Setting global x-coordinate of section
                this.globalAPortyCoord[secNr][coordNr] = pointMotion.ned.east;  // Setting global y-coordinate of section
                this.globalAPortzCoord[secNr][coordNr] = pointMotion.ned.down;  // Setting global z-coordinate of section
            }
        }  
        
        // ----- global dL -----
        // the global DL corresponds to local dx. When having pitch, 
        // the effective width of sections decrease. 
        // Therefore the global version is used instead.
        Double DoubleNumber = Math.floor(this.sectionNumber*0.5);
        int midSectionNumber = DoubleNumber.intValue();
        double deltaX = Math.abs(this.globalAPortxCoord[midSectionNumber+1][0]-this.globalAPortxCoord[midSectionNumber][0]);
        double deltaY = Math.abs(this.globalAPortyCoord[midSectionNumber+1][0]-this.globalAPortyCoord[midSectionNumber][0]);
        double dxScale = Math.sqrt(deltaX*deltaX + deltaY*deltaY)
                /this.dx[midSectionNumber];
        for (int secNr=0; secNr<this.sectionNumber; secNr++) {
            this.globaldL[secNr] = this.dx[secNr]*dxScale;
        }
    }
     
    
    private void initializeArrays() {
        
        coordNumber = new int[sectionNumber];
        lowestCoordNr = new int[sectionNumber];
        
        localxCoord = new double[sectionNumber];
        localyCoord = new double[sectionNumber][minSections];
        localzCoord = new double[sectionNumber][minSections];
        
        globalStarBxCoord = new double[sectionNumber][minSections];
        globalStarByCoord = new double[sectionNumber][minSections];
        globalStarBzCoord = new double[sectionNumber][minSections];  
        globalAPortxCoord = new double[sectionNumber][minSections];
        globalAPortyCoord = new double[sectionNumber][minSections];
        globalAPortzCoord = new double[sectionNumber][minSections];
        globaldL  = new double[sectionNumber];
        
        firstWetCoordStarb = new int[sectionNumber]; 
        firstWetLengthStarb = new double[sectionNumber];
        
        firstWetCoordAPort = new int[sectionNumber]; 
        firstWetLengthAPort = new double[sectionNumber];
        
        dx = new double[sectionNumber];
        elementLength = new double[sectionNumber][minSections-1];
    } 
    
    
    private void readmgfFile(String filePath) {
        System.out.println("SESFederate: Reading hull geometry file: " + filePath);

        // Local variables

        try {
            BufferedReader input = new BufferedReader(new FileReader(filePath));
            
            // Read toptext. This text is not of interest
            try{
                for (int i=0; i<4; i++){
                    String dummy = input.readLine();
                }
            } catch (EOFException eof){
            }            
            
            // Read lpp
            String dummy = input.readLine();
            dummy = dummy.trim();
            Lpp = Double.valueOf(dummy).doubleValue();
            
            // Read sectionNr
            String sectionDummy = "0";
            try{
                sectionDummy = input.readLine();
            } catch (EOFException eof){
                input.close();
            }

            this.sectionNumber = Integer.parseInt(sectionDummy);
            
            // Initialize variables
            this.initializeArrays();
            
            // Fetch remaining data
            try{
                int i = 0;
                boolean running = true;
                do{
                    if (i!=0)
                        sectionDummy = input.readLine();
                    if (sectionDummy == null)
                        running = false;
                    String xCoordDummy = input.readLine();                   
                    String coordDummy = input.readLine(); 
                    
                    if (coordDummy != null) {
                        int sectionNr = Integer.parseInt(sectionDummy);
                        
                        // Number of coords defining the section
                        coordNumber[sectionNr-1] = Integer.parseInt(coordDummy); 
 
                        // Store the x-coordinate of the section
                        localxCoord[sectionNr-1] = Double.valueOf(xCoordDummy).doubleValue();
                        
                        // Read and store coordinates
                        for (int j=0; j<coordNumber[sectionNr-1]; j++){
                            String yzDummy = input.readLine(); yzDummy = yzDummy.trim();
                            {
                                StringTokenizer dummyToken = new StringTokenizer(yzDummy);
                                localyCoord[sectionNr-1][j] = Double.valueOf(dummyToken.nextToken()).doubleValue();
                                localzCoord[sectionNr-1][j] = Double.valueOf(dummyToken.nextToken()).doubleValue();
                            }
                        }
                        // Find the lowest coordinate z-coordinate on the section
                        // If several has same z-coordinate - fint the one closest to CL
                        i++;
                        int candidate = 0;
                        for (int j=1; j<coordNumber[sectionNr-1]; j++) {
                            if (localzCoord[sectionNr-1][candidate] > localzCoord[sectionNr-1][j])
                                candidate = j;
                            else if (localzCoord[sectionNr-1][candidate] == localzCoord[sectionNr-1][j]
                                    && localyCoord[sectionNr-1][candidate] > localyCoord[sectionNr-1][j])
                                candidate = j;
                        }
                        lowestCoordNr[sectionNr-1] = candidate;
                        i++;
                    }
                } while (running == true);
            } catch (EOFException eof){
                input.close();
            }                                
        } catch (IOException e) {
            System.err.println("Error: " + e.toString());
        }    
        

    
    }
    
    
    private void defineElementLengths() {
        double y1;
        double z1;
        double y2;
        double z2;
        double lengthSecond;
        double length;
        for (int secNr=0; secNr<this.sectionNumber; secNr++) {
            for (int coordNr=1; coordNr<this.coordNumber[secNr]; coordNr++) {
                y1 = this.localyCoord[secNr][coordNr-1];
                z1 = this.localzCoord[secNr][coordNr-1];
                y2 = this.localyCoord[secNr][coordNr];
                z2 = this.localzCoord[secNr][coordNr];

                lengthSecond = (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1);
                length = Math.sqrt(lengthSecond);
                this.elementLength[secNr][coordNr-1] = length;
            }   

            if (secNr == 0) {
                // Stern
                this.dx[secNr] = this.Lpp/2 - 0.5*Math.abs(this.localxCoord[0]+this.localxCoord[1]);
            } else if (secNr == this.sectionNumber-1) {
                // Bow
                this.dx[secNr] = this.Lpp/2 - 0.5*Math.abs(this.localxCoord[secNr]+this.localxCoord[secNr-1]);
            } else {
                // Between stern and bow
                this.dx[secNr] = 0.5*Math.abs(this.localxCoord[secNr+1]-this.localxCoord[secNr-1]);
            }
        }        
    }
}